from tkinter import Canvas, Tk, Toplevel
import datetime
import time
import os
import platform
from PIL import Image, ImageTk
from casymda.CoreBlocks.CoreComponent import CoreComponentVisualizer, States


# this class enables the simple 2D animation of component contents and entity flows
# works with CoreComponents and CoreEntities
# basic components: canvas, text and icons (PhotoImages, created and moved on the canvas)
# TODO: mac os x compatibility (.png transparency issue)

class ProcessVisualizer(CoreComponentVisualizer):
    def __init__(self, flowSpeed=200, master=None, BGPhotoFile='sampleModels/bpmn/core-example.jpg', title="Core Example", defaultEntityIconPath=""):
        if master is None:
            self.gui = Tk()
        else:
            self.gui = Toplevel(master=master)

        bgphoto_image = Image.open(BGPhotoFile).convert("RGBA") if platform.system(
        ) != 'Darwin' else Image.open(BGPhotoFile).convert("RGB")
        self.BGphoto = ImageTk.PhotoImage(bgphoto_image)

        self.guiWidth = self.BGphoto.width()
        self.guiHeight = self.BGphoto.height()

        self.gui.geometry(str(self.guiWidth) + "x" + str(self.guiHeight))
        self.gui.title(title)

        self.canvas = Canvas(self.gui, width=self.guiWidth,
                             height=self.guiHeight, bg='black', )
        self.canvas.pack(side='left')

        self.img = self.canvas.create_image(
            0, 0, image=self.BGphoto, anchor='nw')
        timeText = "Time:  " + str(0)
        self.timeLabel = self.canvas.create_text(self.guiWidth - 5, self.guiHeight - 5, font="Helvetica 10", fill='black',
                                                 anchor='se', text=timeText)

        self.componentAnims = []
        self.entityAnims = []

        self.flowSpeed = flowSpeed  # speed of entity movement in pixels per second
        self.defaultEntityIconPath = defaultEntityIconPath

        self.state_icon_paths = self.get_state_icon_paths()

        self.gui.update()

    def animateComponent(self, component, queuing=True, direction=-20):
        # creates new or updates existing animation
        componentAnim = self.find_or_create_component_anim(component)

        # update component's text
        if len(component.entities) > 5:  # don't provide too much details
            contentListString = "Amount:  " + str(len(component.entities))
        else:
            contentListString = "\n".join([" - ".join([entity.name, str(datetime.timedelta(
                seconds=int(entity.timeOfArrival)))]) for entity in component.entities])
        self.canvas.itemconfig(componentAnim.contentList,
                               text=contentListString)

        self.updateTimeLabel(component.env.now)

        # update entity animations as well if they should queue
        if queuing:
            x, y = component.xy
            q_offset_x = 0
            for entity in component.entities:
                self.animateFlowXY(x, y, x + q_offset_x, y,
                                   entity, skipFlow=True)
                q_offset_x += direction

    def animateFlow(self, fromComp, toComp, entity, skipFlow=False, ways=None):
        # animates entity movements between components

        self.updateTimeLabel(fromComp.env.now)

        if toComp is None:
            self.destroyEntityAnim(entity)  # sink
            return

        fx, fy = fromComp.xy
        tx, ty = toComp.xy

        # if waypoints exist, animate flow between waypoints first and update fx, fy
        if ways is not None and toComp.name in ways:
            for way in fromComp.ways[toComp.name]:
                self.animateFlowXY(fx, fy, *way, entity, skipFlow=skipFlow)
                fx, fy = way

        self.animateFlowXY(fx, fy, tx, ty, entity, skipFlow=skipFlow)

    def animateFlowXY(self, fx, fy, tx, ty, entity, skipFlow):
        # flow animation

        # find / create entity animation
        entityAnim = next(
            (x for x in self.entityAnims if x.entity is entity), None)  # existing
        if entityAnim is None:
            entityAnim = self.createEntityAnim(entity, (fx, fy))  # new
        entityIcon = entityAnim.icon

        # make sure starting position is as expected
        self.canvas.coords(entityIcon, fx, fy)
        self.moveIcon(fx, fy, tx, ty, entityIcon,
                      self.flowSpeed, skipFlow=skipFlow)

        entityAnim.xy = (tx, ty)  # update known position

    def moveIcon(self, fx, fy, tx, ty, icon, flowSpeed, skipFlow):
        # only animate flow if not skipped
        if skipFlow:
            self.canvas.move(icon, (tx - fx), (ty - fy))
            self.gui.update()
            return

        d = ((tx - fx)**2 + (ty - fy)**2)**0.5
        if d == 0:
            return  # target already reached

        t = d / float(flowSpeed)  # wallclock time needed for distance
        fps = 30  # aninmations per second
        # x-distance / number of animations = distance per move
        perMoveX = (tx - fx) / (t * fps)
        perMoveY = (ty - fy) / (t * fps)  # y
        remainingX = tx - fx
        remainingY = ty - fy
        # animation loop
        while (abs(remainingX) + abs(remainingY)) > 0:
            # do not move further than necessary
            x_move = remainingX if abs(
                remainingX) < abs(perMoveX) else perMoveX
            y_move = remainingY if abs(
                remainingY) < abs(perMoveY) else perMoveY

            remainingX -= x_move
            remainingY -= y_move

            self.canvas.move(icon, x_move, y_move)
            self.gui.update()
            time.sleep(1 / float(fps))  # sleep until next frame

    def updateTimeLabel(self, now):
        textString = "Time:  " + str(datetime.timedelta(seconds=int(now)))
        self.canvas.itemconfig(self.timeLabel, text=textString)
        self.gui.update()

    def createEntityAnim(self, entity, xy):
        x, y = xy

        try:
            filePath = entity.processVisualizerAnimPath
        except AttributeError:
            filePath = self.defaultEntityIconPath

        # TODO: solve transparency issue -> mac: either not displayed or just a black square (when converting to RGB)
        photo = ImageTk.PhotoImage(
            Image.open(filePath).convert("RGBA") if platform.system(
            ) != 'Darwin' else Image.open(filePath).convert("RGB")
        )

        icon = self.canvas.create_image(x, y, image=photo, anchor='c')
        self.gui.update()

        entityAnim = EntityAnim(entity, icon, photo, xy)
        self.entityAnims.append(entityAnim)

        return entityAnim

    def destroyEntityAnim(self, entity):
        entityAnim = next(
            (x for x in self.entityAnims if x.entity is entity), None)  # existing
        if entityAnim is not None:
            self.entityAnims.remove(entityAnim)
            entityAnim = None
            self.gui.update()

    def createComponentAnim(self, component):
        # currently only consists of updated list of entities
        if component.xyContentList is None:
            # default contents list's position to slightly below the center
            xC, yC = component.xy
            xC -= 20
            yC += 10
        else:
            xC, yC = component.xyContentList

        contentList = self.canvas.create_text(xC, yC, font="Helvetica 10", fill='black',
                                              anchor='nw', text="contents:\n-")

        componentAnim = ComponentAnim(component, contentList)
        self.componentAnims.append(componentAnim)
        return componentAnim

    def find_or_create_component_anim(self, component):
        # creates new or updates existing animation
        componentAnim = next(
            (x for x in self.componentAnims if x.component is component), None)  # existing
        if componentAnim is None:
            componentAnim = self.createComponentAnim(component)  # new
        return componentAnim

    def change_component_state(self, component, state, is_active):

        component_anim = self.find_or_create_component_anim(component)
        if is_active is True:
            state_icon = State_Icon(
                state.value, component.xy, self.state_icon_paths[state.value], self.canvas)
            component_anim.state_icons[state.value] = state_icon
        else:
            if component_anim.state_icons[state.value] is not None:
                self.canvas.delete(
                    component_anim.state_icons[state.value].icon)
                component_anim.state_icons[state.value] = None
        self.gui.update()

    def get_state_icon_paths(self):
        path = os.path.dirname(os.path.realpath(__file__))
        state_icon_paths = {
            enum_state.value: path + "/state_icons/" + enum_state.value + ".png" for enum_state in States
        }
        return state_icon_paths


class State_Icon():

    OFFSETS = {
        States.busy.value: (38, -25),
        States.empty.value: (38, -25),
        States.blocked.value: (38, 0)
    }

    def __init__(self, state_value, xy, path_to_icon, canvas):
        self.state_value = state_value
        self.xy = map(sum, zip(xy, self.OFFSETS[state_value]))
        self.photo = ImageTk.PhotoImage(
            Image.open(path_to_icon).convert("RGBA") if platform.system(
            ) != 'Darwin' else Image.open(path_to_icon).convert("RGB")
        )
        self.icon = canvas.create_image(
            *self.xy, image=self.photo, anchor='c')


class EntityAnim():
    def __init__(self, entity, icon, photo, xy):
        self.entity = entity
        self.icon = icon
        self.photo = photo
        self.xy = xy


class ComponentAnim():
    def __init__(self, component, contentList, icon=None, text=None):
        self.component = component
        self.contentList = contentList
        self.icon = icon
        self.text = text
        self.state_icons = {state.value: None for state in States}



from casymda.CoreBlocks.CoreComponent import CoreComponent


# sink; yields 0 timeout event before no further processing is started

class CoreSink(CoreComponent):
    def __init__(self, env, name, xy=None, ways={}):
        super().__init__(env, name, xy=xy)

    def processEntity(self, entity):
        yield self.env.timeout(0)

        self.onEnter(entity)
        self.entities.append(entity)
        # TODO: remove release since not required with infinite capacity
        self.component_resource.release(entity.component_resource_request)
        self.onExit(entity, None)



import json

# this class provides the possibility to extract and preserve information from a given model
# the model is expected to consist of CoreComponents
# preserved information: name and entry time of entities for every component ("Work-in-Progress" - WIP)
# information is saved to dict/json
# init_model initializes a model with a given
# TODO: verify wip behavior


def snap_model(model, json_path):
    snap = {}

    snap["time"] = model.env.now

    snap["components"] = {}
    for component in model.modelComponents.values():
        # save contents of every component to json incl. timestamp
        snap["components"][component.name] = {}
        snap["components"][component.name]["entities"] = []
        for entity in component.entities:
            entitySnap = {}
            entitySnap["name"] = entity.name
            entitySnap["timeOfArrival"] = entity.timeOfArrival
            if hasattr(entity, "SPR_processingStartedAt"):
                entitySnap["SPR_processingStartedAt"] = entity.SPR_processingStartedAt
            snap["components"][component.name]["entities"].append(
                entitySnap)

    # save to json
    with open(json_path, "w+") as json_file:
        json.dump(snap, json_file, indent=4)

    print("WIP SNAP AT:   " + str(model.env.now))


def init_model(model_class, json_path, entity_type):
    # returns initialized model incl. environment etc, ready for further processing/execution

    with open(json_path, "r") as json_file:
        snap = json.load(json_file)

    snap_time = snap["time"]
    model = model_class(initial_time=snap_time)

    for component in snap["components"]:
        for entityInfo in snap["components"][component]["entities"]:
            entity = entity_type(model.env, entityInfo["name"])
            entity.timeOfArrival = entityInfo["timeOfArrival"]
            if "SPR_processingStartedAt" in entityInfo:
                entity.SPR_processingStartedAt = entityInfo["SPR_processingStartedAt"]

            model.env.process(
                model.modelComponents[component].wipProcessEntity(entity))

    return model



import simpy
from casymda.CoreBlocks.CoreComponent import CoreComponent


# processed entity waits for interrupt before being forwarded to the successor

class CoreWaitForInterrupt(CoreComponent):
    def __init__(self, env, name, xy=None, ways={}):
        super().__init__(env, name, xy=xy, ways=ways)

    # note: entity.currentProcess might be used for interrupt
    def actualProcessing(self, entity):
        try:
            yield self.env.timeout(float('inf'))
        except simpy.Interrupt:
            pass



from casymda.CoreBlocks.CoreComponent import CoreComponent
from casymda.CoreBlocks.CoreEntity import CoreEntity


# the source has a scheduled "creationloop" process, which creates entities and then processes them regularly
# (as if the entity just entered a normal core-component, but with 0 time delay)

class CoreSource(CoreComponent):
    def __init__(self, env, name, xy=None, ways={}, entityType=CoreEntity, interArrivalTime=0, maxEntities=5):

        super().__init__(env, name, xy, ways=ways)

        self.entityType = entityType
        self.interArrivalTime = interArrivalTime
        self.maxEntities = maxEntities

        self.entityCounter = 0

        self.env.process(self.creationLoop())

    def creationLoop(self):
        # this additional call is added due to WIP init (enables reduction of maxEntities before starting)
        yield self.env.timeout(self.interArrivalTime)
        for _ in range(self.maxEntities):
            self.entityCounter += 1
            entity = self.entityType(
                self.env, "entity_" + str(self.entityCounter))

            entity.component_resource_request = self.component_resource.request()
            yield entity.component_resource_request

            self.env.process(self.processEntity(entity))
            yield self.env.timeout(self.interArrivalTime)

    def actualProcessing(self, entity):
        yield self.env.timeout(0)



from casymda.CoreBlocks.CoreComponent import CoreComponent


# processed entity waits for process time (timeout) before being forwarded to the successor

class CoreProcess(CoreComponent):
    def __init__(self, env, name, xy=None, processTime=1.0, ways={}):
        super().__init__(env, name, xy=xy, ways=ways)
        # set processTime
        self.processTime = processTime

    def actualProcessing(self, entity):
        # calc elapsed processing time (wip init)
        processTime = self.processTime - (self.env.now - entity.timeOfArrival)
        yield self.env.timeout(processTime)



# has environment and name
# to be moved between components, remembering it's last time of arrival
# can seize resources (corresponding requests are stored in list)


class CoreEntity:
    def __init__(self, env, name):
        self.env = env
        self.name = name

        self.seizedResources = []  # requested resources, info needed for release
        self.timeOfArrival = None



# base class for blocks, using simpy environment's processes and time
# has environment, name, successor(s), and some more stuff
# calls visualizer if present
from simpy import Resource
from enum import Enum


class CoreComponent():

    def __init__(self, env, name, xy=None, xyContentList=None, queuing=True, ways={}, skipFlow=False, component_capacity=float('inf')):
        self.env = env  # simpy environment
        self.name = name  # string name of block (supposed to be unique)
        self.xy = xy  # coordinate of block for (entity-) animation
        self.xyContentList = xyContentList  # coordinates of the content list
        self.queuing = queuing  # bool whether visualizer should queue arriving entities
        # optional dict with waypoints to be followed by entity when flowing to successor
        self.ways = ways
        # bool to determine if entity flow to successor should be animated:
        self.skipFlow = skipFlow

        self.successors = []  # blocks to forward entities to
        self.entities = []  # list of currently processed entities
        self.visualizers = []  # list of visualizers to be called on status change
        self.overallCountIn = 0  # count overall number of entered entities
        # list of methods to call on enter (signature: component, entity)
        self.doOnEnterList = []
        # list of methods to call on exit (signature: component, entity, successor)
        self.doOnExitList = []
        self.component_resource = Resource(
            self.env, capacity=component_capacity)
        self.state = ComponentState(self)
        # wait until visualizers are set
        self.env.process(self.late_state_evaluation())

    def animate(self):
        for visualizer in self.visualizers:
            visualizer.animateComponent(self, queuing=self.queuing)

    def animateFlow(self, entity, toComp):
        for visualizer in self.visualizers:
            visualizer.animateFlow(
                self, toComp, entity, skipFlow=self.skipFlow, ways=self.ways)

    def onEnter(self, entity):
        entity.timeOfArrival = self.env.now
        self.overallCountIn += 1
        for method in self.doOnEnterList:
            method(self, entity)

    def onExit(self, entity, successor):
        self.entities.remove(entity)
        for method in self.doOnExitList:
            method(self, entity, successor)
        self.animate()
        self.animateFlow(entity, successor)

    # main entry point for entities coming from predecessors
    def processEntity(self, entity):
        self.onEnter(entity)
        yield self.env.process(self.commonProcessEntity(entity))

    # to be called for wip init of entities which already started processing
    # onEnter (incl. arrival time stamp) is omitted to account for already elapsed processing time
    def wipProcessEntity(self, entity):
        req = self.component_resource.request()
        yield req
        entity.component_resource_request = req
        self.env.process(self.commonProcessEntity(entity))

    # common behavior for both types of init
    def commonProcessEntity(self, entity):
        self.entities.append(entity)
        self.animate()
        # processing
        self.state.increment_state_count(States.busy)
        entity.currentProcess = self.env.process(self.actualProcessing(entity))
        yield entity.currentProcess  # entity.currentProcess might be used for interrupt
        # forwarding
        successor = self.findNextSuccessor(entity)
        req = successor.component_resource.request()
        self.state.increment_state_count(States.blocked)
        self.state.decrement_state_count(States.busy)
        yield req  # wait until successor is able to accept the entity
        self.state.decrement_state_count(States.blocked)
        self.component_resource.release(entity.component_resource_request)
        entity.component_resource_request = req  # remember to be released again
        self.onExit(entity, successor)
        self.env.process(successor.processEntity(entity))

    def late_state_evaluation(self):
        yield self.env.timeout(0)
        self.state.evaluate_state_count()

    def actualProcessing(self, entity):  # abstract: to be implemented by child method
        raise NotImplementedError(
            self.name + ":  actualProcessing of CoreComponent not implemented")

    def findNextSuccessor(self, entity):
        return self.successors[0]


# visualizer methods are defined in the "abstract" visualizer class
class CoreComponentVisualizer():
    def animateComponent(self, component, queuing=None, direction=None):
        raise NotImplementedError()

    def animateFlow(self, fromComp, toComp, entity, skipFlow=None, ways=None):
        raise NotImplementedError()

    def destroyEntityAnim(self, entity):
        raise NotImplementedError


class States(Enum):
    empty = "empty"
    busy = "busy"
    blocked = "blocked"


class ComponentState():

    def __init__(self, component):
        self._state_count = {state.value: 0 for state in [
            States.busy, States.blocked]}
        self._current_states = {state.value: False for state in States}
        self.component = component

    def increment_state_count(self, state):
        self._change_state_count(state, 1)

    def decrement_state_count(self, state):
        self._change_state_count(state, -1)

    def _change_state_count(self, state, change):
        self._state_count[state.value] += change
        self.evaluate_state_count()

    def evaluate_state_count(self):
        for state in [States.busy, States.blocked]:
            if self._state_count[state.value] > 0:
                self._set_current_state(state, True)
            else:
                self._set_current_state(state, False)

        if sum(self._state_count.values()) == 0:
            self._set_current_state(States.empty, True)
        else:
            self._set_current_state(States.empty, False)

    def _set_current_state(self, state, new_value):
        if self._current_states[state.value] != new_value:
            for visualizer in self.component.visualizers:
                visualizer.change_component_state(
                    self.component, state, new_value)
            self._current_states[state.value] = new_value



from casymda.CoreBlocks.CoreComponent import CoreComponent


# release requested resource

class CoreReleaseRes(CoreComponent):
    def __init__(self, env, name, resourceToRelease, xy=None, ways={}):
        super().__init__(env, name, xy, ways=ways)

        self.resourceToRelease = resourceToRelease

    def actualProcessing(self, entity):

        # find request of defined resource in entity requests list
        req = next(
            x for x in entity.seizedResources if x.resource is self.resourceToRelease)

        # is done immediately, but necessary for generator behavior
        yield self.resourceToRelease.release(req)



from casymda.CoreBlocks.CoreSource import CoreSource
from casymda.CoreBlocks.CoreEntity import CoreEntity
from casymda.CoreBlocks.CoreProcess import CoreProcess
from casymda.CoreBlocks.CoreSeizeResourceQueue import CoreSeizeResourceQueue
from casymda.CoreBlocks.CoreReleaseRes import CoreReleaseRes
from casymda.CoreBlocks.CoreSink import CoreSink
from casymda.CoreBlocks.CoreResource import CoreResource
from casymda.CoreBlocks.CoreGateway import CoreGateway
from casymda.CoreBlocks.CoreSPR import CoreSPR
from casymda.CoreBlocks.CoreWaitForInterrupt import CoreWaitForInterrupt



from casymda.CoreBlocks.CoreComponent import CoreComponent


# waiting for free resource before seizing and moving forward

class CoreSeizeResourceQueue(CoreComponent):
    def __init__(self, env, name, resourceToRequest, xy=None, ways={}):
        super().__init__(env, name, xy, ways=ways)

        self.resourceToRequest = resourceToRequest

    def actualProcessing(self, entity):

        req = self.resourceToRequest.request()
        yield req
        entity.seizedResources.append(req)



import xmltodict
import json
from operator import sub

# TODO: resolve need to always align elements > 0,0 in camunda modeler
# TODO: validate current offset behavior
OFFSET_REDUCTION = (5, 3)

# this module reads from a given bpmn/xml file and completes a given skeleton template by initialized blocks and their order


def Parse_BPMN(BPMNPATH, JSONPATH, TEMPLATEPATH, MODELPATH):
    ##########################################################################################################
    # prepare as json-formatted dict and save json for quick manual analysis if needed

    with open(BPMNPATH, 'rb') as f:
        bpmn = xmltodict.parse(f)
        bpmn = json.loads(json.dumps(bpmn, indent=4))

    with open(JSONPATH, 'w+') as f:
        json.dump(bpmn, f, indent=4)

    ##########################################################################################################
    # bpmn parsing

    mapping = {
        "dataStoreReference": "resources",
        "startEvent": "sources",
        "endEvent": "sinks",
        "task": "tasks",
        "exclusiveGateway": "xGateways",
        "sequenceFlow": "connectors",
        "textAnnotation": "textAnnotations",
        "association": "associations"
    }

    elements = {
        value: getElements(bpmn, key) for key, value in mapping.items()
    }

    insertions = []
    # resources
    """
    for resource in elements["resources"]:
        names = resource["@name"].split(":")
        names = [n.strip() for n in names]
        className = names[0]
        idName = names[1]
        insertions.append("self.%s = %s(self.env, '%s', c)\n\n" % (idName,
                                                                   className, idName))
        resource["idName"] = idName
    """

    # components
    components = elements["sources"] + elements["sinks"] + \
        elements["tasks"] + elements["xGateways"]

    for comp in elements["resources"] + components:
        names = comp["@name"].split(":")
        # add removal of possibly leading/trailing blanks
        names = [n.strip() for n in names]
        comp["className"] = names[0]
        comp["idName"] = names[1]

    offset = get_offset(bpmn)

    for comp in elements["resources"] + components:
        className = comp["className"]
        idName = comp["idName"]

        xy = get_shape_xy(comp["@id"], bpmn, offset=offset)

        resName = getRelatedResourceIDName(comp, elements)

        if resName is None:
            insertions.append("self.%s = %s(self.env, '%s', xy=%s" % (idName,
                                                                      className, idName, xy))
        else:
            insertions.append("self.%s = %s(self.env, '%s', self.%s, xy=%s" % (
                idName, className, idName, resName, xy))

        insertions[-1] += getRelatedTextAnnotations(comp, elements)

        insertions[-1] += ", ways=" + \
            str(getWaysToSuccessors(comp, bpmn, elements, components, offset=offset))

        # close bracket and 2x newline
        insertions[-1] += ")\n\n"

    mC = "self.modelComponents = {" + ", ".join(
        ["'" + comp["idName"] + "': self." + comp["idName"] for comp in components]) + "}"

    modelGraphNames = {
        comp["idName"]: getSuccessorNamesList(comp, elements, components) for comp in components if not (getSuccessorNamesList(comp, elements, components) is None)
    }

    ##########################################################################################################
    # file writing

    # read template file
    with open(TEMPLATEPATH, 'r') as f:
        template = f.readlines()

    # find resources+components insertion space
    for idx, line in enumerate(template):
        if "#!resources+components" in line:
            loc = idx + 1
            template.insert(loc, "\n")
            break

    for iLine in insertions:
        loc += 1
        template.insert(loc, "".join([" " for _ in range(8)]) + iLine)

    # find #!model insertion space
    for idx, line in enumerate(template):
        if "#!model" in line:
            loc = idx
            break

    loc += 1
    template.insert(loc, "\n" + "".join([" " for _ in range(8)]) + mC)
    loc += 1
    template.insert(loc, "\n\n" + "".join([" " for _ in range(8)])
                    + "self.modelGraphNames = " + json.dumps(modelGraphNames))

    with open(MODELPATH, 'w+') as f:
        f.writelines(template)

    print("\nBPMN parsed.\n")


""" BPMN parsed """


def get_offset(xmlo):
    offset = [float('inf'), float('inf')]

    coords_list = []

    for s in xmlo["bpmn:definitions"]["bpmndi:BPMNDiagram"]["bpmndi:BPMNPlane"]["bpmndi:BPMNShape"]:
        # find all bounded elements and include their label bounds
        shape_bounds = [s["dc:Bounds"]]
        if "bpmndi:BPMNLabel" in s:
            shape_bounds.append(s["bpmndi:BPMNLabel"]["dc:Bounds"])
        for bounds in shape_bounds:
            # look for global min
            coords_list.append(
                (
                    float(bounds["@x"]),
                    float(bounds["@y"])
                )
            )

    for edge in xmlo["bpmn:definitions"]["bpmndi:BPMNDiagram"]["bpmndi:BPMNPlane"]["bpmndi:BPMNEdge"]:
        for wp in edge["di:waypoint"]:
            coords_list.append((float(wp["@x"]), float(wp["@y"])))

    for coords in coords_list:
        for idx in range(len(offset)):
            if coords[idx] < offset[idx]:
                offset[idx] = int(coords[idx])

    offset = tuple(map(sub, offset, OFFSET_REDUCTION))

    return offset


def get_shape_xy(elid, xmlo, offset=(0, 0)):
    for s in xmlo["bpmn:definitions"]["bpmndi:BPMNDiagram"]["bpmndi:BPMNPlane"]["bpmndi:BPMNShape"]:
        if s["@bpmnElement"] == elid:
            xy = [float(s["dc:Bounds"]["@x"]) + int(float(s["dc:Bounds"]["@width"]) / 2),
                  float(s["dc:Bounds"]["@y"]) + int(float(s["dc:Bounds"]["@height"]) / 2)]
            xy = tuple(map(lambda x: int(x), map(sub, xy, offset)))
            return xy


def getElements(bpmn, elementType):
    subDict = bpmn['bpmn:definitions']['bpmn:process']
    if 'bpmn:' + elementType in subDict:
        elements = subDict['bpmn:' + elementType]
        return elements if isinstance(elements, list) else [elements]
    else:
        return []


def containsResourceRelation(component):
    if ("bpmn:dataInputAssociation" in component) or ("bpmn:dataOutputAssociation" in component):
        return True
    else:
        return False


def findResIDName(bpmnID, elements):
    for res in elements["resources"]:
        if res["@id"] == bpmnID:
            return res["idName"]


def getRelatedResourceIDName(component, elements):
    if "bpmn:dataInputAssociation" in component:
        bpmnID = component["bpmn:dataInputAssociation"]["bpmn:sourceRef"]
        return findResIDName(bpmnID, elements)
    elif "bpmn:dataOutputAssociation" in component:
        bpmnID = component["bpmn:dataOutputAssociation"]["bpmn:targetRef"]
        return findResIDName(bpmnID, elements)
    else:
        return None


def getRelatedTextAnnotations(component, elements):
    associations = [assoc for assoc in elements["associations"]
                    if assoc["@sourceRef"] == component["@id"]]
    # get corresponding textAnnotations
    annotations = []
    for assoc in associations:
        annotations += [annot for annot in elements["textAnnotations"]
                        if annot["@id"] == assoc["@targetRef"]]
    params = []
    for annot in annotations:
        params += [x.strip() for x in annot["bpmn:text"].split(";")]
    text = ""
    if len(params) > 0:
        text = ", " + ", ".join(params)
    return text


def getOutgoingSFs(comp):
    # get list of outgoing sequence flows
    if "bpmn:outgoing" not in comp:
        return None
    outgoingSFs = comp["bpmn:outgoing"]
    outgoingSFs = [outgoingSFs] if not isinstance(
        outgoingSFs, list) else outgoingSFs
    return outgoingSFs


def getSuccessorNamesList(comp, elements, components):
    outgoingSFs = getOutgoingSFs(comp)
    res = []
    if outgoingSFs is not None:
        for SF in outgoingSFs:
            # find SF element -> find targetRef -> append idName
            for con in elements["connectors"]:
                if con["@id"] == SF:
                    targetID = con["@targetRef"]
                    for targetComp in components:
                        if targetComp["@id"] == targetID:
                            res.append(targetComp["idName"])
                            break
                    break
    return res


def getDiagramWaypointsForSFID(SFID, xmlo, offset=(0, 0)):
    waypoints = []
    for s in xmlo["bpmn:definitions"]["bpmndi:BPMNDiagram"]["bpmndi:BPMNPlane"]["bpmndi:BPMNEdge"]:
        if s["@bpmnElement"] == SFID:
            for waypoint in s["di:waypoint"]:
                waypoints.append(
                    tuple(map(sub,
                              (
                                  int(float(waypoint["@x"])),
                                  int(float(waypoint["@y"]))
                              ),
                              offset))
                )
            return waypoints


def getWaysToSuccessors(comp, xmlo, elements, components, offset=(0, 0)):
    results = {}  # {"suc name":[{x: int, y: int}, ...]}
    outgoingSFs = getOutgoingSFs(comp)
    if outgoingSFs is not None:
        for SF in outgoingSFs:
            # find SF element -> find targetRef -> append idName: waypoints
            for con in elements["connectors"]:
                if con["@id"] == SF:
                    targetID = con["@targetRef"]
                    for targetComp in components:
                        if targetComp["@id"] == targetID:
                            waypoints = getDiagramWaypointsForSFID(
                                con["@id"], xmlo, offset=offset)
                            # [{x: int, y: int}]
                            results[targetComp["idName"]] = waypoints
                            break
                    break
    return results



from casymda.CoreBlocks.CoreComponent import CoreComponent
from casymda.CoreBlocks.CoreResource import CoreResource

# TODO: WIP
# TODO: WIP -> order of initialized entities should be according to timeOfArrival to seize resources accodingly
# TODO: real combination of imported blocks
# ("hierarchy"-feature -> WIP snap needs to still access single components? -> that way the standard way of just comparing arrival time can be maintained)


# entities in this block follow the process: request resource, process, release

class CoreSPR(CoreComponent):
    def __init__(self, env, name, xy=None, ways={}, processTime=1.0, numResources=1):
        super().__init__(env, name, xy=xy, ways=ways)

        self.processTime = processTime

        self.resource = CoreResource(
            self.env, 'resource', capacity=numResources)

    def actualProcessing(self, entity):
        # wait for resource
        req = self.resource.request()
        yield req
        entity.seizedResources.append(req)

        if not hasattr(entity, "SPR_processingStartedAt"):
            # dictionary to save start times of processing in corresponding SPR blocks
            entity.SPR_processingStartedAt = {}

        if self.name in entity.SPR_processingStartedAt:
            processTime = self.processTime - \
                (self.env.now - entity.SPR_processingStartedAt[self.name])
        else:
            entity.SPR_processingStartedAt[self.name] = self.env.now
            processTime = self.processTime

        # do the processing
        yield self.env.timeout(processTime)

        # release resource
        req = next(
            x for x in entity.seizedResources if x.resource is self.resource)
        self.resource.release(req)



from casymda.CoreBlocks.CoreComponent import CoreComponent


# chooses a successor for all incoming entities

class CoreGateway(CoreComponent):
    def __init__(self, env, name, xy=None, ways={}):
        super().__init__(env, name, xy=xy, ways=ways)

        self.succGenerator = self.roundRobin()

    def actualProcessing(self, entity):
        yield self.env.timeout(0)

    def findNextSuccessor(self, entity):
        return next(self.succGenerator)

    def roundRobin(self):
        last = -1
        while True:
            last += 1
            last %= len(self.successors)
            yield self.successors[last]



from setuptools import setup
from casymda import __version__

setup(
    name='casymda',
    url='',
    author='FFC',
    author_email='fladdi.mir@gmx.de',
    packages=['casymda.BPMN', 'casymda.CoreBlocks',
              'casymda.Visualization', 'casymda.WIP'],
    install_requires=['simpy', 'xmltodict', 'Pillow'],
    version=__version__,
    zip_safe=False,
    license='MIT',
    description='A simple DES modeling and simulation environment based on simpy: create models with help of Camunda Modeler; run and debug using a 2D animation based on tkinter and PIL.',
    include_package_data=True,
    package_data={
        'casymda.Visualizer': ['state_icons/*.png']
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'
    ],
)



from simpy import Resource


# extends basic simpy resource by name property

class CoreResource(Resource):
    def __init__(self, env, name, capacity=1, xy=None, ways=None):
        super().__init__(env, capacity=capacity)

        self.name = name
